# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3

__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "SWUGenerator"
__summary__ = "SWU Package Generator for SWUpdate"
__uri__ = "https://github.com/sbabic/swugenerator"

__version__ = "0.6"

__author__ = "Stefano Babic"
__email__ = "stefano.babic@swupdate.org"

__license__ = "GPL version 3"
__copyright__ = f"Copyright 2022-2024 {__author__}"
